@@ -4,6 +4,7 @@
 #include "qwaylandinputdevice_p.h"
 
 #include "qwaylandintegration_p.h"
+#include "qwaylandtextinputv3_p.h"
 #include "qwaylandwindow_p.h"
 #include "qwaylandsurface_p.h"
 #include "qwaylandbuffer_p.h"
@@ -25,9 +26,6 @@
 #include "qwaylandshmbackingstore_p.h"
 #include "qwaylandtextinputv1_p.h"
 #include "qwaylandtextinputv2_p.h"
-#if QT_WAYLAND_TEXT_INPUT_V4_WIP
-#include "qwaylandtextinputv4_p.h"
-#endif // QT_WAYLAND_TEXT_INPUT_V4_WIP
 #include "qwaylandtextinputinterface_p.h"
 #include "qwaylandinputcontext_p.h"
 #include "qwaylandinputmethodcontext_p.h"
@@ -63,7 +61,8 @@ static const int MaxTouchPoints = 10;
 QWaylandInputDevice::Keyboard::Keyboard(QWaylandInputDevice *p)
     : mParent(p)
 {
-    mRepeatTimer.callOnTimeout([&]() {
+    init(p->get_keyboard());
+    mRepeatTimer.callOnTimeout(this, [&]() {
         if (!focusWindow()) {
             // We destroyed the keyboard focus surface, but the server didn't get the message yet...
             // or the server didn't send an enter event first.
@@ -109,7 +108,7 @@ bool QWaylandInputDevice::Keyboard::createDefaultKeymap()
 QWaylandInputDevice::Keyboard::~Keyboard()
 {
     if (mFocus)
-        QWindowSystemInterface::handleWindowActivated(nullptr);
+        QWindowSystemInterface::handleFocusWindowChanged(nullptr);
     if (version() >= 3)
         wl_keyboard_release(object());
     else
@@ -118,15 +117,20 @@ QWaylandInputDevice::Keyboard::~Keyboard()
 
 QWaylandWindow *QWaylandInputDevice::Keyboard::focusWindow() const
 {
-    return mFocus ? QWaylandWindow::fromWlSurface(mFocus) : nullptr;
+    return mFocus ? mFocus->waylandWindow() : nullptr;
 }
 
 QWaylandInputDevice::Pointer::Pointer(QWaylandInputDevice *seat)
     : mParent(seat)
 {
+    init(seat->get_pointer());
 #if QT_CONFIG(cursor)
+    if (auto cursorShapeManager = seat->mQDisplay->cursorShapeManager()) {
+        mCursor.shape.reset(new QWaylandCursorShape(cursorShapeManager->get_pointer(object())));
+    }
+
     mCursor.frameTimer.setSingleShot(true);
-    mCursor.frameTimer.callOnTimeout([&]() {
+    mCursor.frameTimer.callOnTimeout(this, [&]() {
         cursorTimerCallback();
     });
 #endif
@@ -149,23 +153,16 @@ QWaylandWindow *QWaylandInputDevice::Pointer::focusWindow() const
 
 class WlCallback : public QtWayland::wl_callback {
 public:
-    explicit WlCallback(::wl_callback *callback, std::function<void(uint32_t)> fn, bool autoDelete = false)
+    explicit WlCallback(::wl_callback *callback, std::function<void(uint32_t)> fn)
         : QtWayland::wl_callback(callback)
         , m_fn(fn)
-        , m_autoDelete(autoDelete)
     {}
     ~WlCallback() override { wl_callback_destroy(object()); }
-    bool done() const { return m_done; }
     void callback_done(uint32_t callback_data) override {
-        m_done = true;
         m_fn(callback_data);
-        if (m_autoDelete)
-            delete this;
     }
 private:
-    bool m_done = false;
     std::function<void(uint32_t)> m_fn;
-    bool m_autoDelete = false;
 };
 
 class CursorSurface : public QWaylandSurface
@@ -179,12 +176,10 @@ public:
                 m_pointer, &QWaylandInputDevice::Pointer::updateCursor);
     }
 
-    void hide()
+    void reset()
     {
-        uint serial = m_pointer->mEnterSerial;
-        Q_ASSERT(serial);
-        m_pointer->set_cursor(serial, nullptr, 0, 0);
         m_setSerial = 0;
+        m_hotspot = QPoint();
     }
 
     // Size and hotspot are in surface coordinates
@@ -285,7 +280,9 @@ void QWaylandInputDevice::Pointer::updateCursor()
     auto shape = seat()->mCursor.shape;
 
     if (shape == Qt::BlankCursor) {
-        getOrCreateCursorSurface()->hide();
+        if (mCursor.surface)
+            mCursor.surface->reset();
+        set_cursor(mEnterSerial, nullptr, 0, 0);
         return;
     }
 
@@ -301,6 +298,14 @@ void QWaylandInputDevice::Pointer::updateCursor()
         return;
     }
 
+    if (mCursor.shape) {
+        if (mCursor.surface) {
+            mCursor.surface->reset();
+        }
+        mCursor.shape->setShape(mEnterSerial, shape);
+        return;
+    }
+
     if (!mCursor.theme || idealCursorScale() != mCursor.themeBufferScale)
         updateCursorTheme();
 
@@ -364,6 +369,7 @@ void QWaylandInputDevice::Pointer::cursorFrameCallback()
 QWaylandInputDevice::Touch::Touch(QWaylandInputDevice *p)
     : mParent(p)
 {
+    init(p->get_touch());
 }
 
 QWaylandInputDevice::Touch::~Touch()
@@ -375,7 +381,7 @@ QWaylandInputDevice::Touch::~Touch()
 }
 
 QWaylandInputDevice::QWaylandInputDevice(QWaylandDisplay *display, int version, uint32_t id)
-    : QtWayland::wl_seat(display->wl_registry(), id, qMin(version, 8))
+    : QtWayland::wl_seat(display->wl_registry(), id, qMin(version, 9))
     , mQDisplay(display)
     , mDisplay(display->wl_display())
 {
@@ -400,10 +406,8 @@ QWaylandInputDevice::QWaylandInputDevice(QWaylandDisplay *display, int version,
     if (mQDisplay->textInputManagerv2())
         mTextInput.reset(new QWaylandTextInputv2(mQDisplay, mQDisplay->textInputManagerv2()->get_text_input(wl_seat())));
 
-#if QT_WAYLAND_TEXT_INPUT_V4_WIP
-    if (mQDisplay->textInputManagerv4())
-        mTextInput.reset(new QWaylandTextInputv4(mQDisplay, mQDisplay->textInputManagerv4()->get_text_input(wl_seat())));
-#endif // QT_WAYLAND_TEXT_INPUT_V4_WIP
+    if (mQDisplay->textInputManagerv3())
+        mTextInput.reset(new QWaylandTextInputv3(mQDisplay, mQDisplay->textInputManagerv3()->get_text_input(wl_seat())));
 
     if (mQDisplay->textInputMethodManager())
         mTextInputMethod.reset(new QWaylandTextInputMethod(mQDisplay, mQDisplay->textInputMethodManager()->get_text_input_method(wl_seat())));
@@ -423,14 +427,12 @@ void QWaylandInputDevice::seat_capabilities(uint32_t caps)
 
     if (caps & WL_SEAT_CAPABILITY_KEYBOARD && !mKeyboard) {
         mKeyboard.reset(createKeyboard(this));
-        mKeyboard->init(get_keyboard());
     } else if (!(caps & WL_SEAT_CAPABILITY_KEYBOARD) && mKeyboard) {
         mKeyboard.reset();
     }
 
     if (caps & WL_SEAT_CAPABILITY_POINTER && !mPointer) {
         mPointer.reset(createPointer(this));
-        mPointer->init(get_pointer());
 
         auto *pointerGestures = mQDisplay->pointerGestures();
         if (pointerGestures) {
@@ -441,9 +443,9 @@ void QWaylandInputDevice::seat_capabilities(uint32_t caps)
                         MaxTouchPoints, 0, QString(), QPointingDeviceUniqueId(), this);
             QWindowSystemInterface::registerInputDevice(mTouchPadDevice);
             mPointerGesturePinch.reset(pointerGestures->createPointerGesturePinch(this));
-            mPointerGesturePinch->init(pointerGestures->get_pinch_gesture(get_pointer()));
+            mPointerGesturePinch->init(pointerGestures->get_pinch_gesture(mPointer->object()));
             mPointerGestureSwipe.reset(pointerGestures->createPointerGestureSwipe(this));
-            mPointerGestureSwipe->init(pointerGestures->get_swipe_gesture(get_pointer()));
+            mPointerGestureSwipe->init(pointerGestures->get_swipe_gesture(mPointer->object()));
         }
     } else if (!(caps & WL_SEAT_CAPABILITY_POINTER) && mPointer) {
         mPointer.reset();
@@ -453,7 +455,6 @@ void QWaylandInputDevice::seat_capabilities(uint32_t caps)
 
     if (caps & WL_SEAT_CAPABILITY_TOUCH && !mTouch) {
         mTouch.reset(createTouch(this));
-        mTouch->init(get_touch());
 
         if (!mTouchDevice) {
             // TODO number of touchpoints, actual name and ID
@@ -468,6 +469,11 @@ void QWaylandInputDevice::seat_capabilities(uint32_t caps)
     }
 }
 
+void QWaylandInputDevice::seat_name(const QString &name)
+{
+    mSeatName = name;
+}
+
 QWaylandInputDevice::Keyboard *QWaylandInputDevice::createKeyboard(QWaylandInputDevice *device)
 {
     return new Keyboard(device);
@@ -707,8 +713,8 @@ public:
 
 void QWaylandInputDevice::Pointer::pointer_leave(uint32_t time, struct wl_surface *surface)
 {
-    releaseButtons();
     invalidateFocus();
+    mButtons = Qt::NoButton;
 
     mParent->mTime = time;
 
@@ -823,8 +829,6 @@ void QWaylandInputDevice::Pointer::pointer_button(uint32_t serial, uint32_t time
     default: return; // invalid button number (as far as Qt is concerned)
     }
 
-    mLastButton = qt_button;
-
     if (state)
         mButtons |= qt_button;
     else
@@ -863,13 +867,10 @@ void QWaylandInputDevice::Pointer::invalidateFocus()
 
 void QWaylandInputDevice::Pointer::releaseButtons()
 {
-    if (mButtons == Qt::NoButton)
-        return;
-
     mButtons = Qt::NoButton;
 
     if (auto *window = focusWindow()) {
-        ReleaseEvent e(focusWindow(), mParent->mTime, mSurfacePos, mGlobalPos, mButtons, mLastButton, mParent->modifiers());
+        ReleaseEvent e(focusWindow(), mParent->mTime, mSurfacePos, mGlobalPos, mButtons, Qt::NoButton, mParent->modifiers());
         window->handleMouse(mParent, e);
     }
 }
@@ -879,9 +880,9 @@ class WheelEvent : public QWaylandPointerEvent
 public:
     WheelEvent(QWaylandWindow *surface, Qt::ScrollPhase phase, ulong timestamp, const QPointF &local,
                const QPointF &global, const QPoint &pixelDelta, const QPoint &angleDelta,
-               Qt::MouseEventSource source, Qt::KeyboardModifiers modifiers)
+               Qt::MouseEventSource source, Qt::KeyboardModifiers modifiers, bool inverted)
         : QWaylandPointerEvent(QEvent::Wheel, phase, surface, timestamp,
-                               local, global, pixelDelta, angleDelta, source, modifiers)
+                               local, global, pixelDelta, angleDelta, source, modifiers, inverted)
     {
     }
 };
@@ -975,8 +976,9 @@ void QWaylandInputDevice::Pointer::pointer_axis_stop(uint32_t time, uint32_t axi
     if (!target)
         target = focusWindow();
     Qt::KeyboardModifiers mods = mParent->modifiers();
+    const bool inverted = mFrameData.verticalAxisInverted || mFrameData.horizontalAxisInverted;
     WheelEvent wheelEvent(focusWindow(), Qt::ScrollEnd, mParent->mTime, mSurfacePos, mGlobalPos,
-                          QPoint(), QPoint(), Qt::MouseEventNotSynthesized, mods);
+                          QPoint(), QPoint(), Qt::MouseEventNotSynthesized, mods, inverted);
     target->handleMouse(mParent, wheelEvent);
     mScrollBeginSent = false;
     mScrollDeltaRemainder = QPointF();
@@ -1025,6 +1027,21 @@ void QWaylandInputDevice::Pointer::pointer_axis_value120(uint32_t axis, int32_t
     }
 }
 
+void QWaylandInputDevice::Pointer::pointer_axis_relative_direction(uint32_t axis, uint32_t direction)
+{
+    const bool inverted = direction == axis_relative_direction_inverted;
+    switch (axis) {
+    case axis_vertical_scroll:
+        mFrameData.verticalAxisInverted = inverted;
+        break;
+    case axis_horizontal_scroll:
+        mFrameData.horizontalAxisInverted = inverted;
+        break;
+    default:
+        qCWarning(lcQpaWaylandInput) << "wl_pointer.axis_relative_direction: Unknown axis:" << axis;
+    }
+}
+
 void QWaylandInputDevice::Pointer::setFrameEvent(QWaylandPointerEvent *event)
 {
     qCDebug(lcQpaWaylandInput) << "Setting frame event " << event->type;
@@ -1046,6 +1063,8 @@ void QWaylandInputDevice::Pointer::FrameData::resetScrollData()
     delta120 = QPoint();
     delta = QPointF();
     axisSource = axis_source_wheel;
+    horizontalAxisInverted = false;
+    verticalAxisInverted = false;
 }
 
 bool QWaylandInputDevice::Pointer::FrameData::hasPixelDelta() const
@@ -1127,7 +1146,7 @@ void QWaylandInputDevice::Pointer::flushScrollEvent()
             target->handleMouse(mParent, WheelEvent(focusWindow(), Qt::ScrollBegin, mParent->mTime,
                                                     mSurfacePos, mGlobalPos, QPoint(), QPoint(),
                                                     Qt::MouseEventNotSynthesized,
-                                                    mParent->modifiers()));
+                                                    mParent->modifiers(), false));
             mScrollBeginSent = true;
             mScrollDeltaRemainder = QPointF();
         }
@@ -1136,11 +1155,15 @@ void QWaylandInputDevice::Pointer::flushScrollEvent()
         QPoint pixelDelta = mFrameData.pixelDeltaAndError(&mScrollDeltaRemainder);
         Qt::MouseEventSource source = mFrameData.wheelEventSource();
 
+
+        // The wayland protocol has separate horizontal and vertical axes, Qt has just the one inverted flag
+        // Pragmatically it should't come up
+        const bool inverted = mFrameData.verticalAxisInverted || mFrameData.horizontalAxisInverted;
+
         qCDebug(lcQpaWaylandInput) << "Flushing scroll event" << phase << pixelDelta << angleDelta;
         target->handleMouse(mParent, WheelEvent(focusWindow(), phase, mParent->mTime, mSurfacePos, mGlobalPos,
-                                                pixelDelta, angleDelta, source, mParent->modifiers()));
+                                                pixelDelta, angleDelta, source, mParent->modifiers(), inverted));
     }
-
     mFrameData.resetScrollData();
 }
 
@@ -1215,13 +1238,17 @@ void QWaylandInputDevice::Keyboard::keyboard_enter(uint32_t time, struct wl_surf
         return;
     }
 
+    QWaylandWindow *window = QWaylandWindow::fromWlSurface(surface);
+    if (!window)
+        return;
+
     if (mFocus) {
         qCWarning(lcQpaWayland()) << "Unexpected wl_keyboard.enter event. Keyboard already has focus";
-        disconnect(focusWindow(), &QWaylandWindow::wlSurfaceDestroyed, this, &Keyboard::handleFocusDestroyed);
+        disconnect(mFocus, &QWaylandSurface::destroyed, this, &Keyboard::handleFocusDestroyed);
     }
 
-    mFocus = surface;
-    connect(focusWindow(), &QWaylandWindow::wlSurfaceDestroyed, this, &Keyboard::handleFocusDestroyed);
+    mFocus = window->waylandSurface();
+    connect(mFocus, &QWaylandSurface::destroyed, this, &Keyboard::handleFocusDestroyed);
 
     mParent->mQDisplay->handleKeyboardFocusChanged(mParent);
 }
@@ -1235,13 +1262,17 @@ void QWaylandInputDevice::Keyboard::keyboard_leave(uint32_t time, struct wl_surf
         return;
     }
 
-    if (surface != mFocus) {
+    QWaylandWindow *window = QWaylandWindow::fromWlSurface(surface);
+    if (!window)
+        return;
+
+    if (window->waylandSurface() != mFocus) {
         qCWarning(lcQpaWayland) << "Ignoring unexpected wl_keyboard.leave event."
                                 << "wl_surface argument does not match the current focus"
                                 << "This is most likely a compositor bug";
         return;
     }
-    disconnect(focusWindow(), &QWaylandWindow::wlSurfaceDestroyed, this, &Keyboard::handleFocusDestroyed);
+    disconnect(mFocus, &QWaylandSurface::destroyed, this, &Keyboard::handleFocusDestroyed);
     handleFocusLost();
 }
 
@@ -1305,8 +1336,7 @@ void QWaylandInputDevice::Keyboard::keyboard_key(uint32_t serial, uint32_t time,
         auto code = key + 8; // map to wl_keyboard::keymap_format::keymap_format_xkb_v1
 
         xkb_keysym_t sym = xkb_state_key_get_one_sym(mXkbState.get(), code);
-
-        Qt::KeyboardModifiers modifiers = mParent->modifiers();
+        Qt::KeyboardModifiers modifiers = QXkbCommon::modifiers(mXkbState.get(), sym);
 
         int qtkey = keysymToQtKey(sym, modifiers, mXkbState.get(), code);
         QString text = QXkbCommon::lookupString(mXkbState.get(), code);
@@ -1341,11 +1371,6 @@ void QWaylandInputDevice::Keyboard::keyboard_key(uint32_t serial, uint32_t time,
 
 void QWaylandInputDevice::Keyboard::handleFocusDestroyed()
 {
-    // The signal is emitted by QWaylandWindow, which is not necessarily destroyed along with the
-    // surface, so we still need to disconnect the signal
-    auto *window = qobject_cast<QWaylandWindow *>(sender());
-    disconnect(window, &QWaylandWindow::wlSurfaceDestroyed, this, &Keyboard::handleFocusDestroyed);
-    Q_ASSERT(window->wlSurface() == mFocus);
     handleFocusLost();
 }
 
@@ -1518,7 +1543,12 @@ void QWaylandInputDevice::Touch::touch_frame()
     QWindow *window = mFocus ? mFocus->window() : nullptr;
 
     if (mFocus) {
-        const QWindowSystemInterface::TouchPoint &tp = mPendingTouchPoints.last();
+        // Returns a reference to the last item in the list. The list must not be empty.
+        // If the list can be empty, call isEmpty() before calling this function.
+        // See: https://doc.qt.io/qt-5.15/qlist.html#last
+        if (mPendingTouchPoints.empty())
+            return;
+        const QWindowSystemInterface::TouchPoint &tp = mPendingTouchPoints.constLast();
         // When the touch event is received, the global pos is calculated with the margins
         // in mind. Now we need to adjust again to get the correct local pos back.
         QMargins margins = window->frameMargins();
